package ru.dbotthepony.kommons.arrays

import java.nio.ByteBuffer
import java.nio.LongBuffer

abstract class Long2DArray : Array2D() {
	abstract operator fun get(column: Int, row: Int): Long
	abstract operator fun set(column: Int, row: Int, value: Long)

	fun storeColumnRow(target: ByteBuffer) = matrixStoreColumnRow(this, target, Long2DArray::get, ByteBuffer::putLong)
	fun storeRowColumn(target: ByteBuffer) = matrixStoreRowColumn(this, target, Long2DArray::get, ByteBuffer::putLong)
	fun storeColumnRow(target: LongBuffer) = matrixStoreColumnRow(this, target, Long2DArray::get, LongBuffer::put)
	fun storeRowColumn(target: LongBuffer) = matrixStoreRowColumn(this, target, Long2DArray::get, LongBuffer::put)
	fun store(target: Long2DArray) { target.load(this) }

	fun loadColumnRow(target: ByteBuffer) = matrixLoadColumnRow(this, target, Long2DArray::set, ByteBuffer::getLong)
	fun loadRowColumn(target: ByteBuffer) = matrixLoadRowColumn(this, target, Long2DArray::set, ByteBuffer::getLong)
	fun loadColumnRow(target: LongBuffer) = matrixLoadColumnRow(this, target, Long2DArray::set, LongBuffer::get)
	fun loadRowColumn(target: LongBuffer) = matrixLoadRowColumn(this, target, Long2DArray::set, LongBuffer::get)
	fun load(from: Long2DArray) { matrixLoad(from, this, Long2DArray::get, Long2DArray::set) }
	fun storeTransposed(from: Long2DArray) { matrixLoadTransposed(from, this, Long2DArray::get, Long2DArray::set) }

	open fun contentEquals(other: Long2DArray): Boolean = matrixContentEquals(this, other) { a, b, c, d -> a[c, d] == b[c, d] }

	open fun add(other: Long2DArray): Long2DArray = matrixPlainOperator(this, other, Long2DArray::get, Long::plus, Long2DArray::set)
	open fun sub(other: Long2DArray): Long2DArray = matrixPlainOperator(this, other, Long2DArray::get, Long::minus, Long2DArray::set)
	open fun add(other: Long): Long2DArray = matrixPlainOperator(this, other, Long2DArray::get, Long::plus, Long2DArray::set)
	open fun sub(other: Long): Long2DArray = matrixPlainOperator(this, other, Long2DArray::get, Long::minus, Long2DArray::set)
	open fun mul(other: Long): Long2DArray = matrixPlainOperator(this, other, Long2DArray::get, Long::times, Long2DArray::set)
	open fun div(other: Long): Long2DArray = matrixPlainOperator(this, other, Long2DArray::get, Long::div, Long2DArray::set)

	/**
	 * Allocates new array, unlike other methods which modify object in place
	 */
	open fun mul(other: Long2DArray): Long2DArray {
		return matrixTimes(this, other, Long2DArray::get, Long2DArray::set, Long::plus, Long::times, ::allocate)
	}

	override fun equals(other: Any?): Boolean {
		return matrixEquals(this, other) { a, b, c, d -> a[c, d] == b[c, d] }
	}

	override fun hashCode(): Int {
		return matrixHashCode(this) { a, b, c -> a[b, c].hashCode() }
	}

	override fun toString(): String {
		return matrixToString(this, Long2DArray::get)
	}

	open val determinant: Long? get() {
		if (!isSquareShaped || columns <= 0)
			return null
		else
			return matrixDeterminant(this)
	}

	open val cofactorMatrix: Long2DArray? get() {
		if (!isSquareShaped || columns <= 0)
			return null

		val new = allocate(columns, rows)
		matrixCofactor(this, new, false)
		return new
	}

	open val adjugateMatrix: Long2DArray? get() {
		if (!isSquareShaped || columns <= 0)
			return null

		val new = allocate(columns, rows)
		matrixCofactor(this, new, true)
		return new
	}

	open val inverted: Long2DArray? get() {
		val determinant = determinant ?: return null
		if (determinant == 0L) return null
		val adj = adjugateMatrix!!
		adj.div(determinant)
		return adj
	}

	open val transposed: Long2DArray get() {
		return allocate(rows, columns).also { it.load(this) }
	}

	private class Impl(
		override val columns: Int,
		override val rows: Int
	) : Long2DArray() {
		private val mem = LongArray(columns * rows)

		override fun get(column: Int, row: Int): Long {
			try {
				return mem[column + row * columns]
			} catch (err: IndexOutOfBoundsException) {
				throw IndexOutOfBoundsException("Matrix index out of bounds: $column x $row, while this array has dimensions of $columns x $rows")
			}
		}

		override fun set(column: Int, row: Int, value: Long) {
			try {
				mem[column + row * columns] = value
			} catch (err: IndexOutOfBoundsException) {
				throw IndexOutOfBoundsException("Matrix index out of bounds: $column x $row, while this array has dimensions of $columns x $rows")
			}
		}

		override fun equals(other: Any?): Boolean {
			if (other is Impl) return sizeEquals(other) && mem.contentEquals(other.mem)
			return super.equals(other)
		}

		override fun toString(): String {
			return "Long2DArray" + super.toString()
		}
	}

	private class View(private val parent: Long2DArray) : Long2DArray() {
		override val rows: Int
			get() = parent.rows
		override val columns: Int
			get() = parent.columns

		override fun get(column: Int, row: Int): Long {
			return parent[column, row]
		}

		override fun set(column: Int, row: Int, value: Long) {
			throw UnsupportedOperationException("Read only view")
		}

		override fun equals(other: Any?): Boolean {
			return other is View && parent == other.parent
		}

		override fun hashCode(): Int {
			return parent.hashCode()
		}

		override fun toString(): String {
			return "View = $parent"
		}
	}

	companion object {
		@JvmStatic
		fun allocate(columns: Int, rows: Int): Long2DArray {
			return Impl(columns, rows)
		}

		@JvmStatic
		fun unmodifiable(parent: Long2DArray): Long2DArray {
			if (parent is View)
				return parent
			else
				return View(parent)
		}

		@JvmField
		val ZERO: Long2DArray = Impl(0, 0)
	}
}
